@@ -6,23 +6,16 @@
               Adrian Herrera <adrian.herrera@anu.edu.au>,
               Michal Zalewski
 
-   LLVM integration design comes from Laszlo Szekeres. C bits copied-and-pasted
-   from afl-as.c are Michal's fault.
-
    NGRAM previous location coverage comes from Adrian Herrera.
 
    Copyright 2015, 2016 Google Inc. All rights reserved.
-   Copyright 2019-2020 AFLplusplus Project. All rights reserved.
+   Copyright 2019-2024 AFLplusplus Project. All rights reserved.
 
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at:
 
-     http://www.apache.org/licenses/LICENSE-2.0
-
-   This library is plugged into LLVM when invoking clang through afl-clang-fast.
-   It tells the compiler to add code roughly equivalent to the bits discussed
-   in ../afl-as.h.
+     https://www.apache.org/licenses/LICENSE-2.0
 
  */
 
@@ -44,22 +37,20 @@
 typedef long double max_align_t;
 #endif
 
-#include "llvm/IR/IRBuilder.h"
-#include "llvm/IR/LegacyPassManager.h"
+#include "llvm/Pass.h"
+#include "llvm/Passes/PassPlugin.h"
+#include "llvm/Passes/PassBuilder.h"
+#include "llvm/IR/PassManager.h"
 #include "llvm/IR/BasicBlock.h"
 #include "llvm/IR/Module.h"
 #include "llvm/Support/Debug.h"
 #include "llvm/Support/MathExtras.h"
-#include "llvm/Transforms/IPO/PassManagerBuilder.h"
+#include "llvm/Passes/OptimizationLevel.h"
 
-#if LLVM_VERSION_MAJOR > 3 || \
-    (LLVM_VERSION_MAJOR == 3 && LLVM_VERSION_MINOR > 4)
-  #include "llvm/IR/DebugInfo.h"
-  #include "llvm/IR/CFG.h"
-#else
-  #include "llvm/DebugInfo.h"
-  #include "llvm/Support/CFG.h"
-#endif
+#include "llvm/IR/DebugInfo.h"
+#include "llvm/IR/CFG.h"
+
+#include "llvm/IR/IRBuilder.h"
 
 #include "afl-llvm-common.h"
 #include "llvm-alternative-coverage.h"
@@ -68,17 +59,16 @@ using namespace llvm;
 
 namespace {
 
-class AFLCoverage : public ModulePass {
+class AFLCoverage : public PassInfoMixin<AFLCoverage> {
 
  public:
-  static char ID;
-  AFLCoverage() : ModulePass(ID) {
+  AFLCoverage() {
 
     initInstrumentList();
 
   }
 
-  bool runOnModule(Module &M) override;
+  PreservedAnalyses run(Module &M, ModuleAnalysisManager &MAM);
 
  protected:
   uint32_t    ngram_size = 0;
@@ -92,33 +82,36 @@ class AFLCoverage : public ModulePass {
 
 }  // namespace
 
-char AFLCoverage::ID = 0;
+extern "C" LLVM_ATTRIBUTE_WEAK PassPluginLibraryInfo llvmGetPassPluginInfo() {
 
-/* needed up to 3.9.0 */
-#if LLVM_VERSION_MAJOR == 3 && \
-    (LLVM_VERSION_MINOR < 9 || \
-     (LLVM_VERSION_MINOR == 9 && LLVM_VERSION_PATCH < 1))
-uint64_t PowerOf2Ceil(unsigned in) {
+  return {LLVM_PLUGIN_API_VERSION, "AFLCoverage", "v0.1",
+          /* lambda to insert our pass into the pass pipeline. */
+          [](PassBuilder &PB) {
 
-  uint64_t in64 = in - 1;
-  in64 |= (in64 >> 1);
-  in64 |= (in64 >> 2);
-  in64 |= (in64 >> 4);
-  in64 |= (in64 >> 8);
-  in64 |= (in64 >> 16);
-  in64 |= (in64 >> 32);
-  return in64 + 1;
+#if LLVM_VERSION_MAJOR >= 16
+            PB.registerOptimizerEarlyEPCallback(
+#else
+            PB.registerOptimizerLastEPCallback(
+#endif
+                [](ModulePassManager &MPM, OptimizationLevel OL
+#if LLVM_VERSION_MAJOR >= 20
+                   ,
+                   ThinOrFullLTOPhase Phase
+#endif
+                ) {
+
+                  MPM.addPass(AFLCoverage());
+
+                });
+
+          }};
 
 }
 
-#endif
-
 /* #if LLVM_VERSION_STRING >= "4.0.1" */
-#if LLVM_VERSION_MAJOR > 4 || \
-    (LLVM_VERSION_MAJOR == 4 && LLVM_VERSION_PATCH >= 1)
-  #define AFL_HAVE_VECTOR_INTRINSICS 1
-#endif
-bool AFLCoverage::runOnModule(Module &M) {
+#define AFL_HAVE_VECTOR_INTRINSICS 1
+
+PreservedAnalyses AFLCoverage::run(Module &M, ModuleAnalysisManager &MAM) {
 
   LLVMContext &C = M.getContext();
 
@@ -144,6 +137,13 @@ bool AFLCoverage::runOnModule(Module &M) {
 
   if (getenv("AFL_DEBUG")) debug = 1;
 
+  if (getenv("AFL_LLVM_ONLY_FSRV")) {
+
+    if (debug) { fprintf(stderr, "Instrumentation disabled\n"); }
+    return PreservedAnalyses::all();
+
+  }
+
   if ((isatty(2) && !getenv("AFL_QUIET")) || getenv("AFL_DEBUG") != NULL) {
 
     SAYF(cCYA "afl-llvm-pass" VERSION cRST
@@ -168,7 +168,7 @@ bool AFLCoverage::runOnModule(Module &M) {
 
   /* Decide instrumentation ratio */
 
-  char *       inst_ratio_str = getenv("AFL_INST_RATIO");
+  char        *inst_ratio_str = getenv("AFL_INST_RATIO");
   unsigned int inst_ratio = 100;
 
   if (inst_ratio_str) {
@@ -179,9 +179,6 @@ bool AFLCoverage::runOnModule(Module &M) {
 
   }
 
-#if LLVM_VERSION_MAJOR < 9
-  char *neverZero_counters_str = getenv("AFL_LLVM_NOT_ZERO");
-#endif
   skip_nozero = getenv("AFL_LLVM_SKIP_NEVERZERO");
   use_threadsafe_counters = getenv("AFL_LLVM_THREADSAFE_INST");
 
@@ -299,38 +296,26 @@ bool AFLCoverage::runOnModule(Module &M) {
 
 #ifdef AFL_HAVE_VECTOR_INTRINSICS
   int PrevLocVecSize = PowerOf2Ceil(PrevLocSize);
-  if (ngram_size)
-    PrevLocTy = VectorType::get(IntLocTy, PrevLocVecSize
-  #if LLVM_VERSION_MAJOR >= 12
-                                ,
-                                false
-  #endif
-    );
+  if (ngram_size) PrevLocTy = VectorType::get(IntLocTy, PrevLocVecSize, false);
 #endif
 
 #ifdef AFL_HAVE_VECTOR_INTRINSICS
   int PrevCallerVecSize = PowerOf2Ceil(PrevCallerSize);
-  if (ctx_k)
-    PrevCallerTy = VectorType::get(IntLocTy, PrevCallerVecSize
-  #if LLVM_VERSION_MAJOR >= 12
-                                   ,
-                                   false
-  #endif
-    );
+  if (ctx_k) PrevCallerTy = VectorType::get(IntLocTy, PrevCallerVecSize, false);
 #endif
 
   /* Get globals for the SHM region and the previous location. Note that
      __afl_prev_loc is thread-local. */
 
   GlobalVariable *AFLMapPtr =
-      new GlobalVariable(M, PointerType::get(Int8Ty, 0), false,
+      new GlobalVariable(M, PointerType::get(C, 0), false,
                          GlobalValue::ExternalLinkage, 0, "__afl_area_ptr");
   GlobalVariable *AFLPrevLoc;
   GlobalVariable *AFLPrevCaller;
   GlobalVariable *AFLContext = NULL;
 
   if (ctx_str || caller_str)
-#if defined(__ANDROID__) || defined(__HAIKU__)
+#if defined(__ANDROID__) || defined(__HAIKU__) || defined(NO_TLS)
     AFLContext = new GlobalVariable(
         M, Int32Ty, false, GlobalValue::ExternalLinkage, 0, "__afl_prev_ctx");
 #else
@@ -341,7 +326,7 @@ bool AFLCoverage::runOnModule(Module &M) {
 
 #ifdef AFL_HAVE_VECTOR_INTRINSICS
   if (ngram_size)
-  #if defined(__ANDROID__) || defined(__HAIKU__)
+  #if defined(__ANDROID__) || defined(__HAIKU__) || defined(NO_TLS)
     AFLPrevLoc = new GlobalVariable(
         M, PrevLocTy, /* isConstant */ false, GlobalValue::ExternalLinkage,
         /* Initializer */ nullptr, "__afl_prev_loc");
@@ -354,7 +339,7 @@ bool AFLCoverage::runOnModule(Module &M) {
   #endif
   else
 #endif
-#if defined(__ANDROID__) || defined(__HAIKU__)
+#if defined(__ANDROID__) || defined(__HAIKU__) || defined(NO_TLS)
     AFLPrevLoc = new GlobalVariable(
         M, Int32Ty, false, GlobalValue::ExternalLinkage, 0, "__afl_prev_loc");
 #else
@@ -365,7 +350,7 @@ bool AFLCoverage::runOnModule(Module &M) {
 
 #ifdef AFL_HAVE_VECTOR_INTRINSICS
   if (ctx_k)
-  #if defined(__ANDROID__) || defined(__HAIKU__)
+  #if defined(__ANDROID__) || defined(__HAIKU__) || defined(NO_TLS)
     AFLPrevCaller = new GlobalVariable(
         M, PrevCallerTy, /* isConstant */ false, GlobalValue::ExternalLinkage,
         /* Initializer */ nullptr, "__afl_prev_caller");
@@ -378,7 +363,7 @@ bool AFLCoverage::runOnModule(Module &M) {
   #endif
   else
 #endif
-#if defined(__ANDROID__) || defined(__HAIKU__)
+#if defined(__ANDROID__) || defined(__HAIKU__) || defined(NO_TLS)
     AFLPrevCaller =
         new GlobalVariable(M, Int32Ty, false, GlobalValue::ExternalLinkage, 0,
                            "__afl_prev_caller");
@@ -403,7 +388,7 @@ bool AFLCoverage::runOnModule(Module &M) {
 
   Constant *PrevLocShuffleMask = ConstantVector::get(PrevLocShuffle);
 
-  Constant *                  PrevCallerShuffleMask = NULL;
+  Constant                   *PrevCallerShuffleMask = NULL;
   SmallVector<Constant *, 32> PrevCallerShuffle = {UndefValue::get(Int32Ty)};
 
   if (ctx_k) {
@@ -423,7 +408,7 @@ bool AFLCoverage::runOnModule(Module &M) {
   // other constants we need
   ConstantInt *One = ConstantInt::get(Int8Ty, 1);
 
-  Value *   PrevCtx = NULL;     // CTX sensitive coverage
+  Value    *PrevCtx = NULL;     // CTX sensitive coverage
   LoadInst *PrevCaller = NULL;  // K-CTX coverage
 
   /* Instrument all the things! */
@@ -438,7 +423,7 @@ bool AFLCoverage::runOnModule(Module &M) {
       fprintf(stderr, "FUNCTION: %s (%zu)\n", F.getName().str().c_str(),
               F.size());
 
-    if (!isInInstrumentList(&F)) { continue; }
+    if (!isInInstrumentList(&F, MNAME)) { continue; }
 
     if (F.size() < function_minimum_size) { continue; }
 
@@ -454,9 +439,13 @@ bool AFLCoverage::runOnModule(Module &M) {
 #ifdef AFL_HAVE_VECTOR_INTRINSICS
         if (ctx_k) {
 
-          PrevCaller = IRB.CreateLoad(AFLPrevCaller);
+          PrevCaller = IRB.CreateLoad(PrevCallerTy, AFLPrevCaller);
           PrevCaller->setMetadata(M.getMDKindID("nosanitize"),
+  #if LLVM_MAJOR >= 20
+                                  MDNode::get(C, {}));
+  #else
                                   MDNode::get(C, None));
+  #endif
           PrevCtx =
               IRB.CreateZExt(IRB.CreateXorReduce(PrevCaller), IRB.getInt32Ty());
 
@@ -465,11 +454,15 @@ bool AFLCoverage::runOnModule(Module &M) {
 #endif
         {
 
-          // load the context ID of the previous function and write to to a
+          // load the context ID of the previous function and write to a
           // local variable on the stack
-          LoadInst *PrevCtxLoad = IRB.CreateLoad(AFLContext);
+          LoadInst *PrevCtxLoad = IRB.CreateLoad(IRB.getInt32Ty(), AFLContext);
           PrevCtxLoad->setMetadata(M.getMDKindID("nosanitize"),
+#if LLVM_MAJOR >= 20
+                                   MDNode::get(C, {}));
+#else
                                    MDNode::get(C, None));
+#endif
           PrevCtx = PrevCtxLoad;
 
         }
@@ -516,7 +509,11 @@ bool AFLCoverage::runOnModule(Module &M) {
             StoreInst *Store =
                 IRB.CreateStore(UpdatedPrevCaller, AFLPrevCaller);
             Store->setMetadata(M.getMDKindID("nosanitize"),
+  #if LLVM_MAJOR >= 20
+                               MDNode::get(C, {}));
+  #else
                                MDNode::get(C, None));
+  #endif
 
           } else
 
@@ -526,7 +523,11 @@ bool AFLCoverage::runOnModule(Module &M) {
             if (ctx_str) NewCtx = IRB.CreateXor(PrevCtx, NewCtx);
             StoreInst *StoreCtx = IRB.CreateStore(NewCtx, AFLContext);
             StoreCtx->setMetadata(M.getMDKindID("nosanitize"),
+#if LLVM_MAJOR >= 20
+                                  MDNode::get(C, {}));
+#else
                                   MDNode::get(C, None));
+#endif
 
           }
 
@@ -541,11 +542,11 @@ bool AFLCoverage::runOnModule(Module &M) {
       // cur_loc++;
       cur_loc = AFL_R(map_size);
 
-/* There is a problem with Ubuntu 18.04 and llvm 6.0 (see issue #63).
-   The inline function successors() is not inlined and also not found at runtime
-   :-( As I am unable to detect Ubuntu18.04 heree, the next best thing is to
-   disable this optional optimization for LLVM 6.0.0 and Linux */
-#if !(LLVM_VERSION_MAJOR == 6 && LLVM_VERSION_MINOR == 0) || !defined __linux__
+      /* There is a problem with Ubuntu 18.04 and llvm 6.0 (see issue #63).
+         The inline function successors() is not inlined and also not found at
+         runtime
+         :-( As I am unable to detect Ubuntu18.04 here, the next best thing is
+         to disable this optional optimization for LLVM 6.0.0 and Linux */
       // only instrument if this basic block is the destination of a previous
       // basic block that has multiple successors
       // this gets rid of ~5-10% of instrumentations that are unnecessary
@@ -590,14 +591,18 @@ bool AFLCoverage::runOnModule(Module &M) {
             IRBuilder<> Post_IRB(Inst);
 
             StoreInst *RestoreCtx;
-  #ifdef AFL_HAVE_VECTOR_INTRINSICS
+#ifdef AFL_HAVE_VECTOR_INTRINSICS
             if (ctx_k)
               RestoreCtx = IRB.CreateStore(PrevCaller, AFLPrevCaller);
             else
-  #endif
+#endif
               RestoreCtx = Post_IRB.CreateStore(PrevCtx, AFLContext);
             RestoreCtx->setMetadata(M.getMDKindID("nosanitize"),
+#if LLVM_MAJOR >= 20
+                                    MDNode::get(C, {}));
+#else
                                     MDNode::get(C, None));
+#endif
 
           }
 
@@ -607,8 +612,6 @@ bool AFLCoverage::runOnModule(Module &M) {
 
       }
 
-#endif
-
       ConstantInt *CurLoc;
 
 #ifdef AFL_HAVE_VECTOR_INTRINSICS
@@ -620,8 +623,24 @@ bool AFLCoverage::runOnModule(Module &M) {
 
       /* Load prev_loc */
 
-      LoadInst *PrevLoc = IRB.CreateLoad(AFLPrevLoc);
-      PrevLoc->setMetadata(M.getMDKindID("nosanitize"), MDNode::get(C, None));
+      LoadInst *PrevLoc;
+
+      if (ngram_size) {
+
+        PrevLoc = IRB.CreateLoad(PrevLocTy, AFLPrevLoc);
+
+      } else {
+
+        PrevLoc = IRB.CreateLoad(IRB.getInt32Ty(), AFLPrevLoc);
+
+      }
+
+      PrevLoc->setMetadata(M.getMDKindID("nosanitize"),
+#if LLVM_MAJOR >= 20
+                           MDNode::get(C, {}));
+#else
+                           MDNode::get(C, None));
+#endif
       Value *PrevLocTrans;
 
 #ifdef AFL_HAVE_VECTOR_INTRINSICS
@@ -644,44 +663,33 @@ bool AFLCoverage::runOnModule(Module &M) {
 
       /* Load SHM pointer */
 
-      LoadInst *MapPtr = IRB.CreateLoad(AFLMapPtr);
-      MapPtr->setMetadata(M.getMDKindID("nosanitize"), MDNode::get(C, None));
+      LoadInst *MapPtr = IRB.CreateLoad(PointerType::get(C, 0), AFLMapPtr);
+      MapPtr->setMetadata(M.getMDKindID("nosanitize"),
+#if LLVM_MAJOR >= 20
+                          MDNode::get(C, {}));
+#else
+                          MDNode::get(C, None));
+#endif
 
       Value *MapPtrIdx;
 #ifdef AFL_HAVE_VECTOR_INTRINSICS
       if (ngram_size)
         MapPtrIdx = IRB.CreateGEP(
-            MapPtr,
+            Int8Ty, MapPtr,
             IRB.CreateZExt(
                 IRB.CreateXor(PrevLocTrans, IRB.CreateZExt(CurLoc, Int32Ty)),
                 Int32Ty));
       else
 #endif
-        MapPtrIdx = IRB.CreateGEP(MapPtr, IRB.CreateXor(PrevLocTrans, CurLoc));
+        MapPtrIdx =
+            IRB.CreateGEP(Int8Ty, MapPtr, IRB.CreateXor(PrevLocTrans, CurLoc));
 
       /* Update bitmap */
 
       if (use_threadsafe_counters) {                              /* Atomic */
-                                     /*
-                                     #if LLVM_VERSION_MAJOR < 9
-                                             if (neverZero_counters_str !=
-                                                 NULL) {  // with llvm 9 we make this the default as the bug
-                                     in llvm
-                                                          // is then fixed
-                                     #else
-                                             if (!skip_nozero) {
-                             
-                                     #endif
-                                               // register MapPtrIdx in a todo list
-                                               todo.push_back(MapPtrIdx);
-                             
-                                             } else {
-                             
-                                     */
+
         IRB.CreateAtomicRMW(llvm::AtomicRMWInst::BinOp::Add, MapPtrIdx, One,
-#if LLVM_VERSION_MAJOR >= 13
                             llvm::MaybeAlign(1),
-#endif
                             llvm::AtomicOrdering::Monotonic);
         /*
 
@@ -691,19 +699,18 @@ bool AFLCoverage::runOnModule(Module &M) {
 
       } else {
 
-        LoadInst *Counter = IRB.CreateLoad(MapPtrIdx);
-        Counter->setMetadata(M.getMDKindID("nosanitize"), MDNode::get(C, None));
+        LoadInst *Counter = IRB.CreateLoad(IRB.getInt8Ty(), MapPtrIdx);
+        Counter->setMetadata(M.getMDKindID("nosanitize"),
+#if LLVM_MAJOR >= 20
+                             MDNode::get(C, {}));
+#else
+                             MDNode::get(C, None));
+#endif
 
         Value *Incr = IRB.CreateAdd(Counter, One);
 
-#if LLVM_VERSION_MAJOR < 9
-        if (neverZero_counters_str !=
-            NULL) {  // with llvm 9 we make this the default as the bug in llvm
-                     // is then fixed
-#else
         if (!skip_nozero) {
 
-#endif
           /* hexcoder: Realize a counter that skips zero during overflow.
            * Once this counter reaches its maximum value, it next increments to
            * 1
@@ -723,7 +730,12 @@ bool AFLCoverage::runOnModule(Module &M) {
         }
 
         IRB.CreateStore(Incr, MapPtrIdx)
-            ->setMetadata(M.getMDKindID("nosanitize"), MDNode::get(C, None));
+            ->setMetadata(M.getMDKindID("nosanitize"),
+#if LLVM_MAJOR >= 20
+                          MDNode::get(C, {}));
+#else
+                          MDNode::get(C, None));
+#endif
 
       }                                                  /* non atomic case */
 
@@ -741,7 +753,12 @@ bool AFLCoverage::runOnModule(Module &M) {
             ShuffledPrevLoc, IRB.CreateLShr(CurLoc, (uint64_t)1), (uint64_t)0);
 
         Store = IRB.CreateStore(UpdatedPrevLoc, AFLPrevLoc);
-        Store->setMetadata(M.getMDKindID("nosanitize"), MDNode::get(C, None));
+        Store->setMetadata(M.getMDKindID("nosanitize"),
+  #if LLVM_MAJOR >= 20
+                           MDNode::get(C, {}));
+  #else
+                           MDNode::get(C, None));
+  #endif
 
       } else
 
@@ -750,7 +767,12 @@ bool AFLCoverage::runOnModule(Module &M) {
 
         Store = IRB.CreateStore(ConstantInt::get(Int32Ty, cur_loc >> 1),
                                 AFLPrevLoc);
-        Store->setMetadata(M.getMDKindID("nosanitize"), MDNode::get(C, None));
+        Store->setMetadata(M.getMDKindID("nosanitize"),
+#if LLVM_MAJOR >= 20
+                           MDNode::get(C, {}));
+#else
+                           MDNode::get(C, None));
+#endif
 
       }
 
@@ -772,7 +794,11 @@ bool AFLCoverage::runOnModule(Module &M) {
 #endif
             RestoreCtx = Post_IRB.CreateStore(PrevCtx, AFLContext);
           RestoreCtx->setMetadata(M.getMDKindID("nosanitize"),
+#if LLVM_MAJOR >= 20
+                                  MDNode::get(C, {}));
+#else
                                   MDNode::get(C, None));
+#endif
 
         }
 
@@ -782,120 +808,6 @@ bool AFLCoverage::runOnModule(Module &M) {
 
     }
 
-#if 0
-    if (use_threadsafe_counters) {                       /*Atomic NeverZero */
-      // handle the list of registered blocks to instrument
-      for (auto val : todo) {
-
-        /* hexcoder: Realize a thread-safe counter that skips zero during
-         * overflow. Once this counter reaches its maximum value, it next
-         * increments to 1
-         *
-         * Instead of
-         * Counter + 1 -> Counter
-         * we inject now this
-         * Counter + 1 -> {Counter, OverflowFlag}
-         * Counter + OverflowFlag -> Counter
-         */
-
-        /* equivalent c code looks like this
-         * Thanks to
-         https://preshing.com/20150402/you-can-do-any-kind-of-atomic-read-modify-write-operation/
-
-            int old = atomic_load_explicit(&Counter, memory_order_relaxed);
-            int new;
-            do {
-
-                 if (old == 255) {
-
-                   new = 1;
-
-                 } else {
-
-                   new = old + 1;
-
-                 }
-
-            } while (!atomic_compare_exchange_weak_explicit(&Counter, &old, new,
-
-         memory_order_relaxed, memory_order_relaxed));
-
-         */
-
-        Value *              MapPtrIdx = val;
-        Instruction *        MapPtrIdxInst = cast<Instruction>(val);
-        BasicBlock::iterator it0(&(*MapPtrIdxInst));
-        ++it0;
-        IRBuilder<> IRB(&(*it0));
-
-        // load the old counter value atomically
-        LoadInst *Counter = IRB.CreateLoad(MapPtrIdx);
-        Counter->setAlignment(llvm::Align());
-        Counter->setAtomic(llvm::AtomicOrdering::Monotonic);
-        Counter->setMetadata(M.getMDKindID("nosanitize"), MDNode::get(C, None));
-
-        BasicBlock *BB = IRB.GetInsertBlock();
-        // insert a basic block with the corpus of a do while loop
-        // the calculation may need to repeat, if atomic compare_exchange is not
-        // successful
-
-        BasicBlock::iterator it(*Counter);
-        it++;  // split after load counter
-        BasicBlock *end_bb = BB->splitBasicBlock(it);
-        end_bb->setName("injected");
-
-        // insert the block before the second half of the split
-        BasicBlock *do_while_bb =
-            BasicBlock::Create(C, "injected", end_bb->getParent(), end_bb);
-
-        // set terminator of BB from target end_bb to target do_while_bb
-        auto term = BB->getTerminator();
-        BranchInst::Create(do_while_bb, BB);
-        term->eraseFromParent();
-
-        // continue to fill instructions into the do_while loop
-        IRB.SetInsertPoint(do_while_bb, do_while_bb->getFirstInsertionPt());
-
-        PHINode *PN = IRB.CreatePHI(Int8Ty, 2);
-
-        // compare with maximum value 0xff
-        auto *Cmp = IRB.CreateICmpEQ(Counter, ConstantInt::get(Int8Ty, -1));
-
-        // increment the counter
-        Value *Incr = IRB.CreateAdd(Counter, One);
-
-        // select the counter value or 1
-        auto *Select = IRB.CreateSelect(Cmp, One, Incr);
-
-        // try to save back the new counter value
-        auto *CmpXchg = IRB.CreateAtomicCmpXchg(
-            MapPtrIdx, PN, Select, llvm::AtomicOrdering::Monotonic,
-            llvm::AtomicOrdering::Monotonic);
-        CmpXchg->setAlignment(llvm::Align());
-        CmpXchg->setWeak(true);
-        CmpXchg->setMetadata(M.getMDKindID("nosanitize"), MDNode::get(C, None));
-
-        // get the result of trying to update the Counter
-        Value *Success =
-            IRB.CreateExtractValue(CmpXchg, ArrayRef<unsigned>({1}));
-        // get the (possibly updated) value of Counter
-        Value *OldVal =
-            IRB.CreateExtractValue(CmpXchg, ArrayRef<unsigned>({0}));
-
-        // initially we use Counter
-        PN->addIncoming(Counter, BB);
-        // on retry, we use the updated value
-        PN->addIncoming(OldVal, do_while_bb);
-
-        // if the cmpXchg was not successful, retry
-        IRB.CreateCondBr(Success, end_bb, do_while_bb);
-
-      }
-
-    }
-
-#endif
-
   }
 
   /*
@@ -939,7 +851,11 @@ bool AFLCoverage::runOnModule(Module &M) {
         ConstantInt *const_loc = ConstantInt::get(Int32Ty, map_size);
         StoreInst *  StoreFinalLoc = IRB.CreateStore(const_loc, AFLFinalLoc);
         StoreFinalLoc->setMetadata(M.getMDKindID("nosanitize"),
-                                     MDNode::get(C, None));
+#if LLVM_MAJOR >= 20
+                                  MDNode::get(C, {}));
+#else
+                                  MDNode::get(C, None));
+#endif
 
       }
 
@@ -956,11 +872,12 @@ bool AFLCoverage::runOnModule(Module &M) {
     else {
 
       char modeline[100];
-      snprintf(modeline, sizeof(modeline), "%s%s%s%s%s",
+      snprintf(modeline, sizeof(modeline), "%s%s%s%s%s%s",
                getenv("AFL_HARDEN") ? "hardened" : "non-hardened",
                getenv("AFL_USE_ASAN") ? ", ASAN" : "",
                getenv("AFL_USE_MSAN") ? ", MSAN" : "",
                getenv("AFL_USE_CFISAN") ? ", CFISAN" : "",
+               getenv("AFL_USE_TSAN") ? ", TSAN" : "",
                getenv("AFL_USE_UBSAN") ? ", UBSAN" : "");
       OKF("Instrumented %d locations (%s mode, ratio %u%%).", inst_blocks,
           modeline, inst_ratio);
@@ -969,20 +886,7 @@ bool AFLCoverage::runOnModule(Module &M) {
 
   }
 
-  return true;
+  return PreservedAnalyses();
 
 }
 
-static void registerAFLPass(const PassManagerBuilder &,
-                            legacy::PassManagerBase &PM) {
-
-  PM.add(new AFLCoverage());
-
-}
-
-static RegisterStandardPasses RegisterAFLPass(
-    PassManagerBuilder::EP_OptimizerLast, registerAFLPass);
-
-static RegisterStandardPasses RegisterAFLPass0(
-    PassManagerBuilder::EP_EnabledOnOptLevel0, registerAFLPass);
-
